@@ -15,23 +15,25 @@
 
 """Functions and classes common to multiple pymongo modules."""
 
-import collections
 import datetime
 import warnings
 
 from bson import SON
 from bson.binary import (STANDARD, PYTHON_LEGACY,
                          JAVA_LEGACY, CSHARP_LEGACY)
-from bson.codec_options import CodecOptions
-from bson.py3compat import string_type, integer_types, iteritems
+from bson.codec_options import CodecOptions, TypeRegistry
+from bson.py3compat import abc, integer_types, iteritems, string_type
 from bson.raw_bson import RawBSONDocument
 from pymongo.auth import MECHANISMS
+from pymongo.compression_support import (validate_compressors,
+                                         validate_zlib_compression_level)
+from pymongo.driver_info import DriverInfo
 from pymongo.errors import ConfigurationError
 from pymongo.monitoring import _validate_event_listeners
 from pymongo.read_concern import ReadConcern
 from pymongo.read_preferences import _MONGOS_MODES, _ServerMode
 from pymongo.ssl_support import validate_cert_reqs
-from pymongo.write_concern import WriteConcern
+from pymongo.write_concern import DEFAULT_WRITE_CONCERN, WriteConcern
 
 try:
     from collections import OrderedDict
@@ -159,10 +161,11 @@ def validate_integer(option, value):
     if isinstance(value, integer_types):
         return value
     elif isinstance(value, string_type):
-        if not value.isdigit():
+        try:
+            return int(value)
+        except ValueError:
             raise ValueError("The value of %s must be "
                              "an integer" % (option,))
-        return int(value)
     raise TypeError("Wrong type for %s, value must be an integer" % (option,))
 
 
@@ -238,13 +241,29 @@ def validate_int_or_basestring(option, value):
     if isinstance(value, integer_types):
         return value
     elif isinstance(value, string_type):
-        if value.isdigit():
+        try:
             return int(value)
-        return value
+        except ValueError:
+            return value
     raise TypeError("Wrong type for %s, value must be an "
                     "integer or a string" % (option,))
 
 
+def validate_non_negative_int_or_basestring(option, value):
+    """Validates that 'value' is an integer or string.
+    """
+    if isinstance(value, integer_types):
+        return value
+    elif isinstance(value, string_type):
+        try:
+            val = int(value)
+        except ValueError:
+            return value
+        return validate_non_negative_integer(option, val)
+    raise TypeError("Wrong type for %s, value must be an "
+                    "non negative integer or a string" % (option,))
+
+
 def validate_positive_float(option, value):
     """Validates that 'value' is a float, or can be converted to one, and is
        positive.
@@ -396,13 +415,21 @@ def validate_auth_mechanism_properties(option, value):
 
 def validate_document_class(option, value):
     """Validate the document_class option."""
-    if not issubclass(value, (collections.MutableMapping, RawBSONDocument)):
+    if not issubclass(value, (abc.MutableMapping, RawBSONDocument)):
         raise TypeError("%s must be dict, bson.son.SON, "
                         "bson.raw_bson.RawBSONDocument, or a "
                         "sublass of collections.MutableMapping" % (option,))
     return value
 
 
+def validate_type_registry(option, value):
+    """Validate the type_registry option."""
+    if value is not None and not isinstance(value, TypeRegistry):
+        raise TypeError("%s must be an instance of %s" % (
+            option, TypeRegistry))
+    return value
+
+
 def validate_list(option, value):
     """Validates that 'value' is a list."""
     if not isinstance(value, list):
@@ -419,7 +446,7 @@ def validate_list_or_none(option, value):
 
 def validate_is_mapping(option, value):
     """Validate the type of method arguments that expect a document."""
-    if not isinstance(value, collections.Mapping):
+    if not isinstance(value, abc.Mapping):
         raise TypeError("%s must be an instance of dict, bson.son.SON, or "
                         "other type that inherits from "
                         "collections.Mapping" % (option,))
@@ -427,7 +454,7 @@ def validate_is_mapping(option, value):
 
 def validate_is_document_type(option, value):
     """Validate the type of method arguments that expect a MongoDB document."""
-    if not isinstance(value, (collections.MutableMapping, RawBSONDocument)):
+    if not isinstance(value, (abc.MutableMapping, RawBSONDocument)):
         raise TypeError("%s must be an instance of dict, bson.son.SON, "
                         "bson.raw_bson.RawBSONDocument, or "
                         "a type that inherits from "
@@ -445,6 +472,24 @@ def validate_appname_or_none(option, value):
     return value
 
 
+def validate_driver_or_none(option, value):
+    """Validate the driver keyword arg."""
+    if value is None:
+        return value
+    if not isinstance(value, DriverInfo):
+        raise TypeError("%s must be an instance of DriverInfo" % (option,))
+    return value
+
+
+def validate_is_callable_or_none(option, value):
+    """Validates that 'value' is a callable."""
+    if value is None:
+        return value
+    if not callable(value):
+        raise ValueError("%s must be a callable" % (option,))
+    return value
+
+
 def validate_ok_for_replace(replacement):
     """Validate a replacement document."""
     validate_is_mapping("replacement", replacement)
@@ -491,9 +536,9 @@ def validate_tzinfo(dummy, value):
 # wtimeoutms is an alias for wtimeout,
 URI_VALIDATORS = {
     'replicaset': validate_string_or_none,
-    'w': validate_int_or_basestring,
-    'wtimeout': validate_integer,
-    'wtimeoutms': validate_integer,
+    'w': validate_non_negative_int_or_basestring,
+    'wtimeout': validate_non_negative_integer,
+    'wtimeoutms': validate_non_negative_integer,
     'fsync': validate_boolean_or_string,
     'j': validate_boolean_or_string,
     'journal': validate_boolean_or_string,
@@ -520,8 +565,11 @@ URI_VALIDATORS = {
     'connect': validate_boolean_or_string,
     'minpoolsize': validate_non_negative_integer,
     'appname': validate_appname_or_none,
+    'driver': validate_driver_or_none,
     'unicode_decode_error_handler': validate_unicode_decode_error_handler,
     'retrywrites': validate_boolean_or_string,
+    'compressors': validate_compressors,
+    'zlibcompressionlevel': validate_zlib_compression_level,
 }
 
 TIMEOUT_VALIDATORS = {
@@ -536,11 +584,13 @@ TIMEOUT_VALIDATORS = {
 
 KW_VALIDATORS = {
     'document_class': validate_document_class,
+    'type_registry': validate_type_registry,
     'read_preference': validate_read_preference,
     'event_listeners': _validate_event_listeners,
     'tzinfo': validate_tzinfo,
     'username': validate_string_or_none,
     'password': validate_string_or_none,
+    'server_selector': validate_is_callable_or_none,
 }
 
 URI_VALIDATORS.update(TIMEOUT_VALIDATORS)
@@ -648,6 +698,14 @@ class BaseObject(object):
         """
         return self.__write_concern
 
+    def _write_concern_for(self, session):
+        """Read only access to the write concern of this instance or session.
+        """
+        # Override this operation's write concern with the transaction's.
+        if session and session._in_transaction:
+            return DEFAULT_WRITE_CONCERN
+        return self.write_concern
+
     @property
     def read_preference(self):
         """Read only access to the read preference of this instance.
@@ -657,6 +715,14 @@ class BaseObject(object):
         """
         return self.__read_preference
 
+    def _read_preference_for(self, session):
+        """Read only access to the read preference of this instance or session.
+        """
+        # Override this operation's read preference with the transaction's.
+        if session:
+            return session._txn_read_preference() or self.__read_preference
+        return self.__read_preference
+
     @property
     def read_concern(self):
         """Read only access to the :class:`~pymongo.read_concern.ReadConcern`
