@@ -1,4 +1,4 @@
-# Copyright 2009-2015 MongoDB, Inc.
+# Copyright 2009-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -34,8 +34,13 @@ from pymongo.errors import (AutoReconnect,
                             InvalidOperation,
                             NotMasterError,
                             OperationFailure)
-from pymongo.message import _CursorAddress, _GetMore, _Query, _convert_exception
-from pymongo.read_preferences import ReadPreference
+from pymongo.message import (_convert_exception,
+                             _CursorAddress,
+                             _GetMore,
+                             _RawBatchGetMore,
+                             _Query,
+                             _RawBatchQuery)
+
 
 _QUERY_OPTIONS = {
     "tailable_cursor": 2,
@@ -45,6 +50,7 @@ _QUERY_OPTIONS = {
     "await_data": 32,
     "exhaust": 64,
     "partial": 128}
+_CURSOR_DOC_FIELDS = {'cursor': {'firstBatch': 1, 'nextBatch': 1}}
 
 
 class CursorType(object):
@@ -101,6 +107,8 @@ class _SocketManager:
 class Cursor(object):
     """A cursor / iterator over Mongo query results.
     """
+    _query_class = _Query
+    _getmore_class = _GetMore
 
     def __init__(self, collection, filter=None, projection=None, skip=0,
                  limit=0, no_cursor_timeout=False,
@@ -109,7 +117,7 @@ class Cursor(object):
                  modifiers=None, batch_size=0, manipulate=True,
                  collation=None, hint=None, max_scan=None, max_time_ms=None,
                  max=None, min=None, return_key=False, show_record_id=False,
-                 snapshot=False, comment=None):
+                 snapshot=False, comment=None, session=None):
         """Create a new cursor.
 
         Should not be called directly by application developers - see
@@ -117,9 +125,19 @@ class Cursor(object):
 
         .. mongodoc:: cursors
         """
+        # Initialize all attributes used in __del__ before possibly raising
+        # an error to avoid attribute errors during garbage collection.
         self.__id = None
         self.__exhaust = False
         self.__exhaust_mgr = None
+        self.__killed = False
+
+        if session:
+            self.__session = session
+            self.__explicit_session = True
+        else:
+            self.__session = None
+            self.__explicit_session = False
 
         spec = filter
         if spec is None:
@@ -192,15 +210,13 @@ class Cursor(object):
         self.__data = deque()
         self.__address = None
         self.__retrieved = 0
-        self.__killed = False
 
         self.__codec_options = collection.codec_options
-        self.__read_preference = collection.read_preference
+        # Read preference is set when the initial find is sent.
+        self.__read_preference = None
         self.__read_concern = collection.read_concern
 
         self.__query_flags = cursor_type
-        if self.__read_preference != ReadPreference.PRIMARY:
-            self.__query_flags |= _QUERY_OPTIONS["slave_okay"]
         if no_cursor_timeout:
             self.__query_flags |= _QUERY_OPTIONS["no_timeout"]
         if allow_partial_results:
@@ -251,9 +267,14 @@ class Cursor(object):
         """
         return self._clone(True)
 
-    def _clone(self, deepcopy=True):
+    def _clone(self, deepcopy=True, base=None):
         """Internal clone helper."""
-        clone = self._clone_base()
+        if not base:
+            if self.__explicit_session:
+                base = self._clone_base(self.__session)
+            else:
+                base = self._clone_base(None)
+
         values_to_clone = ("spec", "projection", "skip", "limit",
                            "max_time_ms", "max_await_time_ms", "comment",
                            "max", "min", "ordering", "explain", "hint",
@@ -263,18 +284,20 @@ class Cursor(object):
                     if k.startswith('_Cursor__') and k[9:] in values_to_clone)
         if deepcopy:
             data = self._deepcopy(data)
-        clone.__dict__.update(data)
-        return clone
+        base.__dict__.update(data)
+        return base
 
-    def _clone_base(self):
+    def _clone_base(self, session):
         """Creates an empty Cursor object for information to be copied into.
         """
-        return Cursor(self.__collection)
+        return self.__class__(self.__collection, session=session)
 
     def __die(self, synchronous=False):
         """Closes this cursor.
         """
-        if self.__id and not self.__killed:
+        already_killed = self.__killed
+        self.__killed = True
+        if self.__id and not already_killed:
             if self.__exhaust and self.__exhaust_mgr:
                 # If this is an exhaust cursor and we haven't completely
                 # exhausted the result set we *must* close the socket
@@ -285,13 +308,16 @@ class Cursor(object):
                     self.__address, self.__collection.full_name)
                 if synchronous:
                     self.__collection.database.client._close_cursor_now(
-                        self.__id, address)
+                        self.__id, address, session=self.__session)
                 else:
-                    self.__collection.database.client.close_cursor(
+                    # The cursor will be closed later in a different session.
+                    self.__collection.database.client._close_cursor(
                         self.__id, address)
         if self.__exhaust and self.__exhaust_mgr:
             self.__exhaust_mgr.close()
-        self.__killed = True
+        if self.__session and not self.__explicit_session:
+            self.__session._end_session(lock=synchronous)
+            self.__session = None
 
     def close(self):
         """Explicitly close / kill this cursor.
@@ -581,6 +607,7 @@ class Cursor(object):
             clone = self.clone()
             clone.skip(index + self.__skip)
             clone.limit(-1)  # use a hard limit
+            clone.__query_flags &= ~CursorType.TAILABLE_AWAIT  # PYTHON-1371
             for doc in clone:
                 return doc
             raise IndexError("no such item for Cursor instance")
@@ -588,7 +615,8 @@ class Cursor(object):
                         "instances" % index)
 
     def max_scan(self, max_scan):
-        """Limit the number of documents to scan when performing the query.
+        """**DEPRECATED** - Limit the number of documents to scan when
+        performing the query.
 
         Raises :class:`~pymongo.errors.InvalidOperation` if this
         cursor has already been used. Only the last :meth:`max_scan`
@@ -596,18 +624,30 @@ class Cursor(object):
 
         :Parameters:
           - `max_scan`: the maximum number of documents to scan
+
+        .. versionchanged:: 3.7
+          Deprecated :meth:`max_scan`. Support for this option is deprecated in
+          MongoDB 4.0. Use :meth:`max_time_ms` instead to limit server side
+          execution time.
         """
         self.__check_okay_to_chain()
         self.__max_scan = max_scan
         return self
 
     def max(self, spec):
-        """Adds `max` operator that specifies upper bound for specific index.
+        """Adds ``max`` operator that specifies upper bound for specific index.
+
+        When using ``max``, :meth:`~hint` should also be configured to ensure
+        the query uses the expected index and starting in MongoDB 4.2
+        :meth:`~hint` will be required.
 
         :Parameters:
           - `spec`: a list of field, limit pairs specifying the exclusive
             upper bound for all keys of a specific index in order.
 
+        .. versionchanged:: 3.8
+           Deprecated cursors that use ``max`` without a :meth:`~hint`.
+
         .. versionadded:: 2.7
         """
         if not isinstance(spec, (list, tuple)):
@@ -618,12 +658,19 @@ class Cursor(object):
         return self
 
     def min(self, spec):
-        """Adds `min` operator that specifies lower bound for specific index.
+        """Adds ``min`` operator that specifies lower bound for specific index.
+
+        When using ``min``, :meth:`~hint` should also be configured to ensure
+        the query uses the expected index and starting in MongoDB 4.2
+        :meth:`~hint` will be required.
 
         :Parameters:
           - `spec`: a list of field, limit pairs specifying the inclusive
             lower bound for all keys of a specific index in order.
 
+        .. versionchanged:: 3.8
+           Deprecated cursors that use ``min`` without a :meth:`~hint`.
+
         .. versionadded:: 2.7
         """
         if not isinstance(spec, (list, tuple)):
@@ -678,7 +725,11 @@ class Cursor(object):
         return self
 
     def count(self, with_limit_and_skip=False):
-        """Get the size of the results set for this query.
+        """**DEPRECATED** - Get the size of the results set for this query.
+
+        The :meth:`count` method is deprecated and **not** supported in a
+        transaction. Please use
+        :meth:`~pymongo.collection.Collection.count_documents` instead.
 
         Returns the number of documents in the results set for this query. Does
         not take :meth:`limit` and :meth:`skip` into account by default - set
@@ -704,16 +755,21 @@ class Cursor(object):
         .. note:: The `with_limit_and_skip` parameter requires server
            version **>= 1.1.4-**
 
+        .. versionchanged:: 3.7
+           Deprecated.
+
         .. versionchanged:: 2.8
            The :meth:`~count` method now supports :meth:`~hint`.
         """
+        warnings.warn("count is deprecated. Use Collection.count_documents "
+                      "instead.", DeprecationWarning, stacklevel=2)
         validate_boolean("with_limit_and_skip", with_limit_and_skip)
         cmd = SON([("count", self.__collection.name),
                    ("query", self.__spec)])
         if self.__max_time_ms is not None:
             cmd["maxTimeMS"] = self.__max_time_ms
         if self.__comment:
-            cmd["$comment"] = self.__comment
+            cmd["comment"] = self.__comment
 
         if self.__hint is not None:
             cmd["hint"] = self.__hint
@@ -724,7 +780,8 @@ class Cursor(object):
             if self.__skip:
                 cmd["skip"] = self.__skip
 
-        return self.__collection._count(cmd, self.__collation)
+        return self.__collection._count(
+            cmd, self.__collation, session=self.__session)
 
     def distinct(self, key):
         """Get a list of distinct values for `key` among all documents
@@ -749,15 +806,23 @@ class Cursor(object):
         if self.__max_time_ms is not None:
             options['maxTimeMS'] = self.__max_time_ms
         if self.__comment:
-            options['$comment'] = self.__comment
+            options['comment'] = self.__comment
         if self.__collation is not None:
             options['collation'] = self.__collation
 
-        return self.__collection.distinct(key, **options)
+        return self.__collection.distinct(
+            key, session=self.__session, **options)
 
     def explain(self):
         """Returns an explain plan record for this cursor.
 
+        .. note:: Starting with MongoDB 3.2 :meth:`explain` uses
+          the default verbosity mode of the `explain command
+          <https://docs.mongodb.com/manual/reference/command/explain/>`_,
+          ``allPlansExecution``. To use a different verbosity use
+          :meth:`~pymongo.database.Database.command` to run the explain
+          command directly.
+
         .. mongodoc:: explain
         """
         c = self.clone()
@@ -784,9 +849,9 @@ class Cursor(object):
         Judicious use of hints can greatly improve query
         performance. When doing a query on multiple fields (at least
         one of which is indexed) pass the indexed field as a hint to
-        the query. Hinting will not do anything if the corresponding
-        index does not exist. Raises
-        :class:`~pymongo.errors.InvalidOperation` if this cursor has
+        the query. Raises :class:`~pymongo.errors.OperationFailure` if the
+        provided hint requires an index that does not exist on this collection,
+        and raises :class:`~pymongo.errors.InvalidOperation` if this cursor has
         already been used.
 
         `index` should be an index as passed to
@@ -812,7 +877,8 @@ class Cursor(object):
         http://docs.mongodb.org/manual/reference/operator/comment/
 
         :Parameters:
-          - `comment`: A string or document
+          - `comment`: A string to attach to the query to help interpret and
+            trace the operation in the server logs and in profile data.
 
         .. versionadded:: 2.7
         """
@@ -878,18 +944,14 @@ class Cursor(object):
         listeners = client._event_listeners
         publish = listeners.enabled_for_commands
         from_command = False
+        start = datetime.datetime.now()
+
+        def duration(): return datetime.datetime.now() - start
 
         if operation:
-            kwargs = {
-                "read_preference": self.__read_preference,
-                "exhaust": self.__exhaust,
-            }
-            if self.__address is not None:
-                kwargs["address"] = self.__address
-
             try:
-                response = client._send_message_with_response(operation,
-                                                              **kwargs)
+                response = client._send_message_with_response(
+                    operation, exhaust=self.__exhaust, address=self.__address)
                 self.__address = response.address
                 if self.__exhaust:
                     # 'response' is an ExhaustResponse.
@@ -897,8 +959,7 @@ class Cursor(object):
                                                         response.pool)
 
                 cmd_name = operation.name
-                data = response.data
-                cmd_duration = response.duration
+                reply = response.data
                 rqst_id = response.request_id
                 from_command = response.from_command
             except AutoReconnect:
@@ -907,6 +968,7 @@ class Cursor(object):
                 # assertion on some server releases if we get here
                 # due to a socket timeout.
                 self.__killed = True
+                self.__die()
                 raise
         else:
             # Exhaust cursor - no getMore message.
@@ -922,29 +984,30 @@ class Cursor(object):
                     cmd['maxTimeMS'] = self.__max_time_ms
                 listeners.publish_command_start(
                     cmd, self.__collection.database.name, 0, self.__address)
-                start = datetime.datetime.now()
             try:
-                data = self.__exhaust_mgr.sock.receive_message(1, None)
+                reply = self.__exhaust_mgr.sock.receive_message(None)
             except Exception as exc:
                 if publish:
-                    duration = datetime.datetime.now() - start
                     listeners.publish_command_failure(
-                        duration, _convert_exception(exc), cmd_name, rqst_id,
+                        duration(), _convert_exception(exc), cmd_name, rqst_id,
                         self.__address)
                 if isinstance(exc, ConnectionFailure):
                     self.__die()
                 raise
-            if publish:
-                cmd_duration = datetime.datetime.now() - start
 
-        if publish:
-            start = datetime.datetime.now()
         try:
-            doc = helpers._unpack_response(response=data,
-                                           cursor_id=self.__id,
-                                           codec_options=self.__codec_options)
+            user_fields = None
+            legacy_response = True
             if from_command:
-                helpers._check_command_response(doc['data'][0])
+                user_fields = _CURSOR_DOC_FIELDS
+                legacy_response = False
+            docs = self._unpack_response(
+                reply, self.__id, self.__collection.codec_options,
+                legacy_response=legacy_response, user_fields=user_fields)
+            if from_command:
+                first = docs[0]
+                client._receive_cluster_time(first, self.__session)
+                helpers._check_command_response(first)
         except OperationFailure as exc:
             self.__killed = True
 
@@ -952,9 +1015,8 @@ class Cursor(object):
             self.__die()
 
             if publish:
-                duration = (datetime.datetime.now() - start) + cmd_duration
                 listeners.publish_command_failure(
-                    duration, exc.details, cmd_name, rqst_id, self.__address)
+                    duration(), exc.details, cmd_name, rqst_id, self.__address)
 
             # If this is a tailable cursor the error is likely
             # due to capped collection roll over. Setting
@@ -972,63 +1034,74 @@ class Cursor(object):
             self.__die()
 
             if publish:
-                duration = (datetime.datetime.now() - start) + cmd_duration
                 listeners.publish_command_failure(
-                    duration, exc.details, cmd_name, rqst_id, self.__address)
+                    duration(), exc.details, cmd_name, rqst_id, self.__address)
 
             client._reset_server_and_request_check(self.__address)
             raise
         except Exception as exc:
             if publish:
-                duration = (datetime.datetime.now() - start) + cmd_duration
                 listeners.publish_command_failure(
-                    duration, _convert_exception(exc), cmd_name, rqst_id,
+                    duration(), _convert_exception(exc), cmd_name, rqst_id,
                     self.__address)
             raise
 
         if publish:
-            duration = (datetime.datetime.now() - start) + cmd_duration
             # Must publish in find / getMore / explain command response format.
             if from_command:
-                res = doc['data'][0]
+                res = docs[0]
             elif cmd_name == "explain":
-                res = doc["data"][0] if doc["number_returned"] else {}
+                res = docs[0] if docs else {}
             else:
-                res = {"cursor": {"id": doc["cursor_id"],
+                res = {"cursor": {"id": reply.cursor_id,
                                   "ns": self.__collection.full_name},
                        "ok": 1}
                 if cmd_name == "find":
-                    res["cursor"]["firstBatch"] = doc["data"]
+                    res["cursor"]["firstBatch"] = docs
                 else:
-                    res["cursor"]["nextBatch"] = doc["data"]
+                    res["cursor"]["nextBatch"] = docs
             listeners.publish_command_success(
-                duration, res, cmd_name, rqst_id, self.__address)
+                duration(), res, cmd_name, rqst_id, self.__address)
 
-        if from_command and cmd_name != "explain":
-            cursor = doc['data'][0]['cursor']
-            self.__id = cursor['id']
-            if cmd_name == 'find':
-                documents = cursor['firstBatch']
+        if from_command:
+            if cmd_name != "explain":
+                cursor = docs[0]['cursor']
+                self.__id = cursor['id']
+                if cmd_name == 'find':
+                    documents = cursor['firstBatch']
+                else:
+                    documents = cursor['nextBatch']
+                self.__data = deque(documents)
+                self.__retrieved += len(documents)
             else:
-                documents = cursor['nextBatch']
-            self.__data = deque(documents)
-            self.__retrieved += len(documents)
+                self.__id = 0
+                self.__data = deque(docs)
+                self.__retrieved += len(docs)
         else:
-            self.__id = doc["cursor_id"]
-            self.__data = deque(doc["data"])
-            self.__retrieved += doc["number_returned"]
+            self.__id = reply.cursor_id
+            self.__data = deque(docs)
+            self.__retrieved += reply.number_returned
 
         if self.__id == 0:
             self.__killed = True
-
+            # Don't wait for garbage collection to call __del__, return the
+            # socket and the session to the pool now.
+            self.__die()
 
         if self.__limit and self.__id and self.__limit <= self.__retrieved:
             self.__die()
 
-        # Don't wait for garbage collection to call __del__, return the
-        # socket to the pool now.
-        if self.__exhaust and self.__id == 0:
-            self.__exhaust_mgr.close()
+    def _unpack_response(self, response, cursor_id, codec_options,
+                         user_fields=None, legacy_response=False):
+        return response.unpack_response(cursor_id, codec_options, user_fields,
+                                        legacy_response)
+
+    def _read_preference(self):
+        if self.__read_preference is None:
+            # Save the read preference for getMore commands.
+            self.__read_preference = self.__collection._read_preference_for(
+                self.session)
+        return self.__read_preference
 
     def _refresh(self):
         """Refreshes the cursor with more data from Mongo.
@@ -1040,21 +1113,31 @@ class Cursor(object):
         if len(self.__data) or self.__killed:
             return len(self.__data)
 
+        if not self.__session:
+            self.__session = self.__collection.database.client._ensure_session()
+
         if self.__id is None:  # Query
-            self.__send_message(_Query(self.__query_flags,
-                                       self.__collection.database.name,
-                                       self.__collection.name,
-                                       self.__skip,
-                                       self.__query_spec(),
-                                       self.__projection,
-                                       self.__codec_options,
-                                       self.__read_preference,
-                                       self.__limit,
-                                       self.__batch_size,
-                                       self.__read_concern,
-                                       self.__collation))
-            if not self.__id:
-                self.__killed = True
+            if (self.__min or self.__max) and not self.__hint:
+                warnings.warn("using a min/max query operator without "
+                              "specifying a Cursor.hint is deprecated. A "
+                              "hint will be required when using min/max in "
+                              "PyMongo 4.0",
+                              DeprecationWarning, stacklevel=3)
+            q = self._query_class(self.__query_flags,
+                                  self.__collection.database.name,
+                                  self.__collection.name,
+                                  self.__skip,
+                                  self.__query_spec(),
+                                  self.__projection,
+                                  self.__codec_options,
+                                  self._read_preference(),
+                                  self.__limit,
+                                  self.__batch_size,
+                                  self.__read_concern,
+                                  self.__collation,
+                                  self.__session,
+                                  self.__collection.database.client)
+            self.__send_message(q)
         elif self.__id:  # Get More
             if self.__limit:
                 limit = self.__limit - self.__retrieved
@@ -1067,15 +1150,16 @@ class Cursor(object):
             if self.__exhaust:
                 self.__send_message(None)
             else:
-                self.__send_message(_GetMore(self.__collection.database.name,
-                                             self.__collection.name,
-                                             limit,
-                                             self.__id,
-                                             self.__codec_options,
-                                             self.__max_await_time_ms))
-
-        else:  # Cursor id is zero nothing else to return
-            self.__killed = True
+                g = self._getmore_class(self.__collection.database.name,
+                                        self.__collection.name,
+                                        limit,
+                                        self.__id,
+                                        self.__codec_options,
+                                        self._read_preference(),
+                                        self.__session,
+                                        self.__collection.database.client,
+                                        self.__max_await_time_ms)
+                self.__send_message(g)
 
         return len(self.__data)
 
@@ -1122,6 +1206,15 @@ class Cursor(object):
         """
         return self.__address
 
+    @property
+    def session(self):
+        """The cursor's :class:`~pymongo.client_session.ClientSession`, or None.
+
+        .. versionadded:: 3.6
+        """
+        if self.__explicit_session:
+            return self.__session
+
     def __iter__(self):
         return self
 
@@ -1192,3 +1285,43 @@ class Cursor(object):
                     key = copy.deepcopy(key, memo)
                 y[key] = value
         return y
+
+
+class RawBatchCursor(Cursor):
+    """A cursor / iterator over raw batches of BSON data from a query result."""
+
+    _query_class = _RawBatchQuery
+    _getmore_class = _RawBatchGetMore
+
+    def __init__(self, *args, **kwargs):
+        """Create a new cursor / iterator over raw batches of BSON data.
+
+        Should not be called directly by application developers -
+        see :meth:`~pymongo.collection.Collection.find_raw_batches`
+        instead.
+
+        .. mongodoc:: cursors
+        """
+        manipulate = kwargs.get('manipulate')
+        kwargs['manipulate'] = False
+        super(RawBatchCursor, self).__init__(*args, **kwargs)
+
+        # Throw only after cursor's initialized, to prevent errors in __del__.
+        if manipulate:
+            raise InvalidOperation(
+                "Cannot use RawBatchCursor with manipulate=True")
+
+    def _unpack_response(self, response, cursor_id, codec_options,
+                         user_fields=None, legacy_response=False):
+        return response.raw_response(cursor_id)
+
+    def explain(self):
+        """Returns an explain plan record for this cursor.
+
+        .. mongodoc:: explain
+        """
+        clone = self._clone(deepcopy=True, base=Cursor(self.collection))
+        return clone.explain()
+
+    def __getitem__(self, index):
+        raise InvalidOperation("Cannot call __getitem__ on RawBatchCursor")
