@@ -48,7 +48,7 @@ _PY37PLUS = sys.version_info[:2] >= (3, 7)
 
 if HAVE_SSL:
     try:
-        # Python 2.7.9+, 3.2+, PyPy 2.5.1+, etc.
+        # Python 2.7.9+, PyPy 2.5.1+, etc.
         from ssl import SSLContext
     except ImportError:
         from pymongo.ssl_context import SSLContext
@@ -123,11 +123,13 @@ if HAVE_SSL:
             # up to date versions of MongoDB 2.4 and above already disable
             # SSLv2 and SSLv3, python disables SSLv2 by default in >= 2.7.7
             # and >= 3.3.4 and SSLv3 in >= 3.4.3. There is no way for us to do
-            # any of this explicitly for python 2.6 or 2.7 before 2.7.9.
+            # any of this explicitly for python 2.7 before 2.7.9.
             ctx.options |= getattr(ssl, "OP_NO_SSLv2", 0)
             ctx.options |= getattr(ssl, "OP_NO_SSLv3", 0)
             # OpenSSL >= 1.0.0
             ctx.options |= getattr(ssl, "OP_NO_COMPRESSION", 0)
+            # Python 3.7+ with OpenSSL >= 1.1.0h
+            ctx.options |= getattr(ssl, "OP_NO_RENEGOTIATION", 0)
         if certfile is not None:
             try:
                 if passphrase is not None:
